@@ -108,6 +108,12 @@ struct _TerminalApp
   GMenu *menubar_new_terminal_section;
   GMenu *menubar_set_profile_section;
   GMenu *menubar_set_encoding_submenu;
+
+  GMenuModel *profilemenu;
+  GMenuModel *headermenu;
+  GMenu *headermenu_set_profile_section;
+  GMenu *headermenu_set_encoding_submenu;
+
   GMenu *set_profile_menu;
 
   GtkClipboard *clipboard;
@@ -432,6 +438,27 @@ append_new_terminal_item (GMenu *section,
   g_menu_append_item (section, item);
 }
 
+static void
+fill_header_new_terminal_menu (GMenuModel *menu,
+                               ProfileData *data,
+                               guint n_profiles)
+{
+  gs_unref_object GMenu *section = NULL;
+
+  if (n_profiles <= 1)
+    return;
+
+  section = g_menu_new ();
+
+  for (guint i = 0; i < n_profiles; i++) {
+    menu_append_numbered (section, data[i].label, i + 1,
+                          "win.new-terminal",
+                          g_variant_new ("(ss)", "default", data[i].uuid));
+  }
+
+  g_menu_append_section (G_MENU (menu), _("New Terminal"), G_MENU_MODEL (section));
+}
+
 static void
 fill_new_terminal_section (TerminalApp *app,
                            GMenu *section,
@@ -467,8 +494,6 @@ set_profile_submenu_new (ProfileData *data,
 static void
 terminal_app_update_profile_menus (TerminalApp *app)
 {
-  g_menu_remove_all (G_MENU (app->menubar_new_terminal_section));
-  g_menu_remove_all (G_MENU (app->menubar_set_profile_section));
   g_clear_object (&app->set_profile_menu);
 
   /* Get profiles list and sort by label */
@@ -485,14 +510,90 @@ terminal_app_update_profile_menus (TerminalApp *app)
   ProfileData *profiles = (ProfileData*) array->data;
   guint n_profiles = array->len;
 
-  fill_new_terminal_section (app, app->menubar_new_terminal_section, profiles, n_profiles);
-
   app->set_profile_menu = set_profile_submenu_new (profiles, n_profiles);
 
-  if (app->set_profile_menu != NULL) {
-    g_menu_append_submenu (app->menubar_set_profile_section, _("Change _Profile"),
-                           G_MENU_MODEL (app->set_profile_menu));
+  if (app->menubar != NULL) {
+    g_menu_remove_all (G_MENU (app->menubar_new_terminal_section));
+    fill_new_terminal_section (app, app->menubar_new_terminal_section, profiles, n_profiles);
+
+    g_menu_remove_all (G_MENU (app->menubar_set_profile_section));
+    if (app->set_profile_menu != NULL) {
+      g_menu_append_submenu (app->menubar_set_profile_section, _("Change _Profile"),
+                             G_MENU_MODEL (app->set_profile_menu));
+    }
   }
+
+  if (app->profilemenu != NULL) {
+    g_menu_remove_all (G_MENU (app->profilemenu));
+    fill_header_new_terminal_menu (app->profilemenu, profiles, n_profiles);
+  }
+
+  if (app->headermenu != NULL) {
+    g_menu_remove_all (G_MENU (app->headermenu_set_profile_section));
+    if (app->set_profile_menu != NULL) {
+      g_menu_append_submenu (app->headermenu_set_profile_section, _("_Profile"),
+                             G_MENU_MODEL (app->set_profile_menu));
+    }
+  }
+}
+
+static GMenuModel *
+terminal_app_create_menubar (TerminalApp *app,
+                             gboolean shell_shows_menubar)
+{
+  /* If the menubar is shown by the shell, omit mnemonics for the submenus. This is because Alt+F etc.
+   * are more important to be usable in the terminal, the menu cannot be replaced runtime (to toggle
+   * between mnemonic and non-mnemonic versions), gtk-enable-mnemonics or gtk_window_set_mnemonic_modifier()
+   * don't effect the menubar either, so there wouldn't be a way to disable Alt+F for File etc. otherwise.
+   * Furthermore, the menu would even grab mnemonics from the File and Preferences windows.
+   * In Unity, Alt+F10 opens the menubar, this should be good enough for keyboard navigation.
+   * If the menubar is shown by the app, toggling mnemonics is handled in terminal-window.c using
+   * gtk_window_set_mnemonic_modifier().
+   * See bug 792978 for details. */
+  terminal_util_load_objects_resource (shell_shows_menubar ? "/org/gnome/terminal/ui/menubar-without-mnemonics.ui"
+                                                           : "/org/gnome/terminal/ui/menubar-with-mnemonics.ui",
+                                       "menubar", &app->menubar,
+                                       "new-terminal-section", &app->menubar_new_terminal_section,
+                                       "set-profile-section", &app->menubar_set_profile_section,
+                                       "set-encoding-submenu", &app->menubar_set_encoding_submenu,
+                                       NULL);
+
+  /* Install the encodings submenu */
+  terminal_encodings_append_menu (app->menubar_set_encoding_submenu);
+
+  /* Install profile sections */
+  terminal_app_update_profile_menus (app);
+
+  return app->menubar;
+}
+
+static void
+terminal_app_create_headermenu (TerminalApp *app)
+{
+  terminal_util_load_objects_resource ("/org/gnome/terminal/ui/headerbar-menu.ui",
+                                       "headermenu", &app->headermenu,
+                                       "set-profile-section", &app->headermenu_set_profile_section,
+#if 0
+                                       "set-encoding-submenu", &app->headermenu_set_encoding_submenu,
+#endif
+                                       NULL);
+
+#if 0
+  /* Install the encodings submenu */
+  terminal_encodings_append_menu (app->headermenu_set_encoding_submenu);
+#endif
+
+  /* Install profile sections */
+  terminal_app_update_profile_menus (app);
+}
+
+static void
+terminal_app_create_profilemenu (TerminalApp *app)
+{
+  app->profilemenu = G_MENU_MODEL (g_menu_new ());
+
+  /* Install profile sections */
+  terminal_app_update_profile_menus (app);
 }
 
 /* Clipboard */
@@ -645,35 +746,17 @@ terminal_app_startup (GApplication *application)
                 "gtk-shell-shows-menubar", &shell_shows_menubar,
                 NULL);
 
-  /* Menubar */
-  /* If the menubar is shown by the shell, omit mnemonics for the submenus. This is because Alt+F etc.
-   * are more important to be usable in the terminal, the menu cannot be replaced runtime (to toggle
-   * between mnemonic and non-mnemonic versions), gtk-enable-mnemonics or gtk_window_set_mnemonic_modifier()
-   * don't effect the menubar either, so there wouldn't be a way to disable Alt+F for File etc. otherwise.
-   * Furthermore, the menu would even grab mnemonics from the File and Preferences windows.
-   * In Unity, Alt+F10 opens the menubar, this should be good enough for keyboard navigation.
-   * If the menubar is shown by the app, toggling mnemonics is handled in terminal-window.c using
-   * gtk_window_set_mnemonic_modifier().
-   * See bug 792978 for details. */
-  terminal_util_load_objects_resource (shell_shows_menubar ? "/org/gnome/terminal/ui/menubar-without-mnemonics.ui"
-                                                           : "/org/gnome/terminal/ui/menubar-with-mnemonics.ui",
-                                       "menubar", &app->menubar,
-                                       "new-terminal-section", &app->menubar_new_terminal_section,
-                                       "set-profile-section", &app->menubar_set_profile_section,
-                                       "set-encoding-submenu", &app->menubar_set_encoding_submenu,
-                                       NULL);
+  /* Create menubar */
+  terminal_app_create_menubar (app, shell_shows_menubar);
 
-  /* Create dynamic menus and keep them updated */
-  terminal_app_update_profile_menus (app);
+  /* Keep dynamic menus updated */
   g_signal_connect_swapped (app->profiles_list, "children-changed",
                             G_CALLBACK (terminal_app_update_profile_menus), app);
 
-  /* Install the encodings submenu */
-  terminal_encodings_append_menu (app->menubar_set_encoding_submenu);
-
   /* Show/hide the menubar as appropriate: If the shell wants to show the menubar, make it available. */
   if (shell_shows_menubar)
-    gtk_application_set_menubar (GTK_APPLICATION (app), app->menubar);
+    gtk_application_set_menubar (GTK_APPLICATION (app),
+                                 terminal_app_get_menubar (app));
 
   _terminal_debug_print (TERMINAL_DEBUG_SERVER, "Startup complete\n");
 }
@@ -705,7 +788,12 @@ terminal_app_init (TerminalApp *app)
    * to override, so we cache them on startup and don't react to changes.
    */
   app->unified_menu = g_settings_get_boolean (app->global_settings, TERMINAL_SETTING_UNIFIED_MENU_KEY);
-  app->use_headerbar = g_settings_get_boolean (app->global_settings, TERMINAL_SETTING_HEADERBAR_KEY);
+
+  const char *env = g_getenv("GNOME_TERMINAL_HEADERBAR");
+  if (env != NULL)
+    app->use_headerbar = g_strcmp0 (env, "1") == 0;
+  else
+    app->use_headerbar = g_settings_get_boolean (app->global_settings, TERMINAL_SETTING_HEADERBAR_KEY);
 
 #if GTK_CHECK_VERSION (3, 19, 0)
   GtkSettings *gtk_settings = gtk_settings_get_default ();
@@ -733,7 +821,7 @@ terminal_app_init (TerminalApp *app)
   app->screen_map = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
 
   gs_unref_object GSettings *settings = g_settings_get_child (app->global_settings, "keybindings");
-  terminal_accels_init (G_APPLICATION (app), settings);
+  terminal_accels_init (G_APPLICATION (app), settings, app->use_headerbar);
 }
 
 static void
@@ -760,6 +848,10 @@ terminal_app_finalize (GObject *object)
   g_clear_object (&app->menubar_new_terminal_section);
   g_clear_object (&app->menubar_set_profile_section);
   g_clear_object (&app->menubar_set_encoding_submenu);
+  g_clear_object (&app->profilemenu);
+  g_clear_object (&app->headermenu);
+  g_clear_object (&app->headermenu_set_profile_section);
+  g_clear_object (&app->headermenu_set_encoding_submenu);
   g_clear_object (&app->set_profile_menu);
 
   terminal_accels_shutdown ();
@@ -1073,6 +1165,36 @@ terminal_app_get_menubar (TerminalApp *app)
   return app->menubar;
 }
 
+/**
+ * terminal_app_get_headermenu:
+ * @app: a #TerminalApp
+ *
+ * Returns: (tranfer none): the main window headerbar menu bar as a #GMenuModel
+ */
+GMenuModel *
+terminal_app_get_headermenu (TerminalApp *app)
+{
+  if (app->headermenu == NULL)
+    terminal_app_create_headermenu (app);
+
+  return app->headermenu;
+}
+
+/**
+ * terminal_app_get_profilemenu:
+ * @app: a #TerminalApp
+ *
+ * Returns: (tranfer none): the main window headerbar profile menu as a #GMenuModel
+ */
+GMenuModel *
+terminal_app_get_profilemenu (TerminalApp *app)
+{
+  if (app->profilemenu == NULL)
+    terminal_app_create_profilemenu (app);
+
+  return app->profilemenu;
+}
+
 /**
  * terminal_app_get_profile_section:
  * @app: a #TerminalApp
@@ -1172,3 +1294,16 @@ terminal_app_get_use_headerbar (TerminalApp *app)
 
   return app->use_headerbar;
 }
+
+gboolean
+terminal_app_get_dialog_use_headerbar (TerminalApp *app)
+{
+  g_return_val_if_fail (TERMINAL_IS_APP (app), FALSE);
+
+  gboolean dialog_use_header;
+  g_object_get (gtk_settings_get_default (),
+                "gtk-dialogs-use-header", &dialog_use_header,
+                NULL);
+
+  return dialog_use_header && app->use_headerbar;
+}
