@@ -751,6 +751,7 @@ terminal_prefs_show_preferences (GSettings *profile, const char *widget_name)
   GtkWidget *theme_variant_label, *theme_variant_combo;
   GtkWidget *new_terminal_mode_label, *new_terminal_mode_combo;
   GtkWidget *close_button, *help_button;
+  GtkWidget *content_box, *general_frame, *keybindings_frame;
   GSettings *settings;
 
   const GActionEntry action_entries[] = {
@@ -771,6 +772,9 @@ terminal_prefs_show_preferences (GSettings *profile, const char *widget_name)
   data->builder = terminal_util_load_widgets_resource ("/org/gnome/terminal/ui/preferences.ui",
                                        "preferences-dialog",
                                        "preferences-dialog", &dialog,
+                                       "dialogue-content-box", &content_box,
+                                       "general-frame", &general_frame,
+                                       "keybindings-frame", &keybindings_frame,
                                        "close-button", &close_button,
                                        "help-button", &help_button,
                                        "default-show-menubar-checkbutton", &show_menubar_button,
@@ -848,7 +852,8 @@ terminal_prefs_show_preferences (GSettings *profile, const char *widget_name)
   gtk_widget_set_visible (theme_variant_combo, FALSE);
 #endif /* GTK+ 3.19 */
 
-  if (terminal_app_get_menu_unified (app)) {
+  if (terminal_app_get_menu_unified (app) ||
+      terminal_app_get_use_headerbar (app)) {
     g_settings_bind (settings,
                      TERMINAL_SETTING_NEW_TERMINAL_MODE_KEY,
                      new_terminal_mode_combo,
@@ -891,6 +896,34 @@ terminal_prefs_show_preferences (GSettings *profile, const char *widget_name)
 
   profile_prefs_init ();
 
+  /* Move action widgets to titlebar when headerbar is used */
+  if (terminal_app_get_dialog_use_headerbar (app)) {
+    GtkWidget *headerbar;
+    GtkWidget *bbox;
+
+    headerbar = g_object_new (GTK_TYPE_HEADER_BAR,
+                              "show-close-button", TRUE,
+                              NULL);
+    bbox = gtk_widget_get_parent (help_button);
+
+    gtk_container_remove (GTK_CONTAINER (bbox), g_object_ref (help_button));
+    gtk_header_bar_pack_start (GTK_HEADER_BAR (headerbar), help_button);
+    g_object_unref (help_button);
+
+    gtk_style_context_add_class (gtk_widget_get_style_context (help_button),
+                                 "text-button");
+
+    gtk_widget_show (headerbar);
+    gtk_widget_hide (bbox);
+
+    gtk_window_set_titlebar (GTK_WINDOW (dialog), headerbar);
+
+    /* Remove extra spacing around the content, and extra frames */
+    g_object_set (G_OBJECT (content_box), "margin", 0, NULL);
+    gtk_frame_set_shadow_type (GTK_FRAME (general_frame), GTK_SHADOW_NONE);
+    gtk_frame_set_shadow_type (GTK_FRAME (keybindings_frame), GTK_SHADOW_NONE);
+  }
+
   /* misc */
 
   g_signal_connect (close_button, "clicked", G_CALLBACK (prefs_dialog_close_button_clicked_cb), data);
